"""Entity linking
This step links the entity mentions in the question to the entities in the Wikidata knowledge graph.
It inference on the REL endpoint.
"""
import argparse
from pathlib import Path

import srsly
from tqdm import tqdm

from .entity_linking import WikidataLinker, DBpediaLinker
from .utils import socket_reachable


def link(args):
    """Link the entities in the questions to the Wikidata knowledge graph"""
    if not socket_reachable(args.el_endpoint):
        raise RuntimeError(f"Can't reach the endpoint {args.el_endpoint}")

    if args.knowledge_graph == 'wikidata':
        linker = WikidataLinker(args.el_endpoint, args.wikimapper_db)
    elif args.knowledge_graph == 'dbpedia':
        linker = DBpediaLinker(args.el_endpoint)
    else:
        raise NotImplementedError(f"Knowledge graph {args.knowledge_graph} not implemented")

    with open(args.input, 'r', encoding='utf-8') as f:
        total_lines = len(f.readlines())
    questions = srsly.read_jsonl(args.input)
    cnt_id_not_found = 0
    cnt_id_found = 0
    all_linked = []
    for question in tqdm(questions, total=total_lines, desc=f"Entity linking {args.input}"):
        linked = linker.annotate(question[args.ground_on])
        cnt_id_found += len(linked["question_entities"])
        if 'not_converted_entities' in linked:
            cnt_id_not_found += len(linked['not_converted_entities'])
        if 'id' in question:
            linked['id'] = question['id']
        all_linked.append(linked)
    if cnt_id_not_found > 0:
        print(f"{cnt_id_not_found} / {cnt_id_found + cnt_id_not_found} grounded entities not converted to ids")
    # check whether the folder exists
    folder_path = Path(args.output).parent
    if not folder_path.exists():
        folder_path.mkdir(parents=True)
        print(f"Folder {folder_path} created")
    Path(args.output).parent.mkdir(parents=True, exist_ok=True)
    srsly.write_jsonl(args.output, all_linked)
    print(f"Entity linking result saved to {args.output}")


def add_arguments(parser):
    """Add entity linking arguments to the parser"""
    parser.description = '''Entity linking on Wikidata.
    The input is a jsonl file. The field of interest is specified by the argument --ground-on.
    The output is a jsonl file, each line is a dict with keys: id, question_entities, spans, entity_names.
    '''
    parser.add_argument('-i', '--input', type=str, help='Input file path, in which the question is stored')
    parser.add_argument('-o', '--output', type=str, help='Output file path, in which the entity linking result is stored')
    parser.add_argument('-e', '--el-endpoint', type=str, default='http://127.0.0.1:1235', help='Entity linking endpoint \
                        (default: http://127.0.0.1:1235 <local REL endpoint>)')
    parser.add_argument('-kg', '--knowledge-graph', type=str, default='wikidata', choices=['wikidata', 'dbpedia'],
                        help='Knowledge graph to link to, only wikidata is supported now')
    parser.add_argument('--wikimapper-db', type=str, default='resources/wikimapper/index_enwiki.db', help='Wikimapper database path')
    parser.add_argument('--ground-on', type=str, default='question', help='The key to ground on, the corresponding text will be sent to the REL endpoint for entity linking')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    link(args)
